package v6_test

import (
	"errors"

	"code.cloudfoundry.org/cli/actor/actionerror"
	"code.cloudfoundry.org/cli/actor/v2action"
	"code.cloudfoundry.org/cli/api/cloudcontroller/ccv2"
	"code.cloudfoundry.org/cli/api/cloudcontroller/ccv2/constant"
	"code.cloudfoundry.org/cli/api/cloudcontroller/ccversion"
	"code.cloudfoundry.org/cli/command/commandfakes"
	. "code.cloudfoundry.org/cli/command/v6"
	"code.cloudfoundry.org/cli/command/v6/v6fakes"
	"code.cloudfoundry.org/cli/util/configv3"
	"code.cloudfoundry.org/cli/util/ui"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	. "github.com/onsi/gomega/gbytes"
)

var _ = Describe("services Command", func() {
	var (
		cmd             ServicesCommand
		testUI          *ui.UI
		fakeConfig      *commandfakes.FakeConfig
		fakeSharedActor *commandfakes.FakeSharedActor
		fakeActor       *v6fakes.FakeServiceInstancesActor
		binaryName      string
		executeErr      error
	)

	BeforeEach(func() {
		testUI = ui.NewTestUI(nil, NewBuffer(), NewBuffer())
		fakeConfig = new(commandfakes.FakeConfig)
		fakeSharedActor = new(commandfakes.FakeSharedActor)
		fakeActor = new(v6fakes.FakeServiceInstancesActor)

		cmd = ServicesCommand{
			UI:          testUI,
			Config:      fakeConfig,
			SharedActor: fakeSharedActor,
			Actor:       fakeActor,
		}

		binaryName = "faceman"
		fakeConfig.BinaryNameReturns(binaryName)

		fakeActor.CloudControllerAPIVersionReturns(ccversion.MinVersionMaintenanceInfoInSummaryV2)
	})

	JustBeforeEach(func() {
		executeErr = cmd.Execute(nil)
	})

	When("an error is encountered checking if the environment is setup correctly", func() {
		BeforeEach(func() {
			fakeSharedActor.CheckTargetReturns(actionerror.NotLoggedInError{BinaryName: binaryName})
		})

		It("returns an error", func() {
			Expect(executeErr).To(MatchError(actionerror.NotLoggedInError{BinaryName: binaryName}))
			Expect(fakeSharedActor.CheckTargetCallCount()).To(Equal(1))
			checkTargetedOrgArg, checkTargetedSpaceArg := fakeSharedActor.CheckTargetArgsForCall(0)
			Expect(checkTargetedOrgArg).To(BeTrue())
			Expect(checkTargetedSpaceArg).To(BeTrue())
		})
	})

	When("the user is logged in and an org and space are targeted", func() {
		BeforeEach(func() {
			fakeConfig.TargetedOrganizationReturns(configv3.Organization{
				Name: "some-org",
			})
			fakeConfig.TargetedSpaceReturns(configv3.Space{
				GUID: "some-space-guid",
				Name: "some-space",
			})
		})

		When("getting the current user fails", func() {
			var expectedErr error

			BeforeEach(func() {
				expectedErr = errors.New("some-error that happened")
				fakeConfig.CurrentUserReturns(configv3.User{}, expectedErr)
			})

			It("returns the error", func() {
				Expect(executeErr).To(MatchError(expectedErr))
				Expect(fakeConfig.CurrentUserCallCount()).To(Equal(1))
			})
		})

		When("getting the current user succeeds", func() {
			var (
				fakeUser configv3.User
			)

			BeforeEach(func() {
				fakeUser = configv3.User{Name: "some-user"}
				fakeConfig.CurrentUserReturns(fakeUser, nil)
				fakeConfig.TargetedSpaceReturns(configv3.Space{
					GUID: "some-space-guid",
					Name: "some-space",
				})
				fakeConfig.TargetedOrganizationReturns(configv3.Organization{
					Name: "some-org",
				})
			})

			When("there are no services", func() {
				BeforeEach(func() {
					fakeActor.GetServiceInstancesSummaryBySpaceReturns(
						nil,
						v2action.Warnings{"get-summary-warnings"},
						nil,
					)
				})

				It("displays that there are no services", func() {
					Expect(executeErr).ToNot(HaveOccurred())
					Expect(testUI.Out).To(Say("Getting services in org %s / space %s as %s...", "some-org",
						"some-space", fakeUser.Name))

					out := testUI.Out.(*Buffer).Contents()
					Expect(out).To(MatchRegexp("No services found"))
					Expect(out).ToNot(MatchRegexp(`name\s+service\s+plan\s+bound apps\s+last operation`))
					Expect(testUI.Err).To(Say("get-summary-warnings"))
				})
			})

			When("there are services", func() {
				BeforeEach(func() {
					fakeActor.GetServiceInstancesSummaryBySpaceReturns(
						[]v2action.ServiceInstanceSummary{
							{
								ServiceInstance: v2action.ServiceInstance{
									Name: "instance-1",
									LastOperation: ccv2.LastOperation{
										Type:  "some-type",
										State: "some-state",
									},
									Type: constant.ManagedService,
									MaintenanceInfo: ccv2.MaintenanceInfo{
										Version: "2.0.0",
									},
								},
								ServicePlan: v2action.ServicePlan{
									Name: "some-plan",
									MaintenanceInfo: ccv2.MaintenanceInfo{
										Version: "3.0.0",
									},
								},
								Service: v2action.Service{
									Label:             "some-service-1",
									ServiceBrokerName: "broker-1",
								},
								BoundApplications: []v2action.BoundApplication{
									{AppName: "app-2"},
									{AppName: "app-1"},
								},
							},
							{
								ServiceInstance: v2action.ServiceInstance{
									Name: "instance-3",
									Type: constant.UserProvidedService,
								},
							},
							{
								ServiceInstance: v2action.ServiceInstance{
									Name: "instance-2",
									Type: constant.ManagedService,
									MaintenanceInfo: ccv2.MaintenanceInfo{
										Version: "2.0.0",
									},
								},
								ServicePlan: v2action.ServicePlan{
									Name: "some-plan",
									MaintenanceInfo: ccv2.MaintenanceInfo{
										Version: "2.0.0",
									},
								},
								Service: v2action.Service{
									Label:             "some-service-2",
									ServiceBrokerName: "broker-2",
								},
							},
						},
						v2action.Warnings{"get-summary-warnings"},
						nil,
					)
				})

				It("displays all the services and apps in alphanumeric sorted order together with the org & space & warnings & upgrades", func() {
					Expect(executeErr).ToNot(HaveOccurred())
					Expect(testUI.Out).To(Say("Getting services in org %s / space %s as %s...", "some-org", "some-space", fakeUser.Name))
					Expect(testUI.Out).To(Say(`name\s+service\s+plan\s+bound apps\s+last operation\s+broker\s+upgrade available`))
					Expect(testUI.Out).To(Say(`instance-1\s+some-service-1\s+some-plan\s+app-1, app-2\s+some-type some-state\s+broker-1\s+yes`))
					Expect(testUI.Out).To(Say(`instance-2\s+some-service-2\s+some-plan\s+broker-2\s+no`))
					Expect(testUI.Out).To(Say(`instance-3\s+user-provided\s+$`))
					Expect(testUI.Err).To(Say("get-summary-warnings"))
				})

				When("CC version is too old to show upgrade information", func() {
					BeforeEach(func() {
						fakeActor.CloudControllerAPIVersionReturns(ccversion.MinSupportedV2ClientVersion)
					})

					It("shows a tip with a minimum service upgrade version required", func() {
						Expect(testUI.Out).To(Say("TIP: Please upgrade to CC API v%s or higher for individual service upgrades", ccversion.MinVersionMaintenanceInfoInSummaryV2))
					})
				})
			})
		})
	})
})
